// Contains class definitions
// Copyright (C) 1997 by John P. Weiss
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the Artistic License, included as the file
// "LICENSE" in the source code archive.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of the file "LICENSE", containing
// the License John Weiss originally placed this program under.
#include "TiKanDB.h"
#include <iomanip.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdio.h>
#include <errno.h>




// ***************
//  Here, we define the actual functions


//------------------------------------------------------------
// Internals.
//------------------------------------------------------------


// function prototype.
//
char *getEntry(char *ibuf, char *rectag, bool verbose=true,
			char delim='=');


// The definition:  getEntry looks to see if "rectag" is at the
// beginning of the string in "ibuf".  If so, it returns a pointer to
// the remainder of the string.  The flag "verbose" indicates whether
// or not to print error messages when it can't find the tag.
//
char *getEntry(char *ibuf, char *rectag, bool verbose, char delim)
{
   char *outptr;
   int matches;
   const char *errmsg="TiKanDB::readEntry() - bad record.";
   
#ifdef DEBUG_ON
   cerr << "%dbg%\tgetEntry(): " << ibuf << endl << flush;
#endif
   outptr = strchr(ibuf,delim);
   if(outptr == NULL)
	 {
	 if(verbose)
	    {
	    cerr << errmsg << endl;
	    cerr << "Couldn't find delimiter '" 
		    << delim << "'." << endl;
	    }
	 return NULL;
	 }

   outptr[0] = '\0';
   outptr++;
   matches = strcmp(ibuf,rectag);
   if(matches != 0)
	 {
	 if(verbose)
	    {
	    cerr << errmsg << endl;
	    cerr << "The '" << rectag << "' entry comes next, not '"
		    << ibuf << "'." << endl;
	    }
	 return NULL;
	 }
   
   return outptr;
}


// Recursively count the files under the given path.
//
int count_files(char *path)
{
   DIR *dp;
   char nbuf[IOBUFLEN];
   struct stat sbuf;
   struct dirent *dbuf;
   int out,ok;
   static int recurs=0;

#ifdef DEBUG_ON
   cerr << "%dbg%\tcount_files(): level "
	   << recurs << endl;
#endif
   out = 0;
   dp = opendir(path);
   if(dp == NULL)
	 {
	 perror("count_files()");
	 return 0;
	 }

   recurs++;
   dbuf = readdir(dp);
   while(dbuf != NULL)
	 {
	 sprintf(nbuf,"%s/%s",path,dbuf->d_name);
#ifdef DEBUG_ON
	 cerr << "\tcount_files(): " << nbuf << endl;
#endif
	 // Check for hidden files [e.g. '.', '..', etc....]
	 if(dbuf->d_name[0]=='.')
	    ok = 0;
	 else
	    ok = stat(nbuf,&sbuf) + 1;  //typ==0 on success, -1 on err.
	 // Check file type.
	 if(S_ISREG(sbuf.st_mode) && ok)
	    {
	    out++;
	    }
	 else if( (S_ISDIR(sbuf.st_mode) || S_ISLNK(sbuf.st_mode))
			&& ok)			
	    {
	    out += count_files(nbuf);
	    }
	 //
	 dbuf = readdir(dp);
	 }
   closedir(dp);
   recurs--;
#ifdef DEBUG_ON
	 cerr << "\tcounted=" << out << endl;
#endif
   return out;
}


// Recursively find all of the regular files under the given path.
// "len" is the size of the list, "listbuf".  The routine returns the
// number of filenames read.
//
int list_files(char *path, char **listbuf, int len)
{
   DIR *dp;
   char nbuf[IOBUFLEN];
   struct stat sbuf;
   struct dirent *dbuf;
   int ok,count,i;
   static int recurs=0;

#ifdef DEBUG_ON
   cerr << "%dbg%\tlist_files(): level "
	   << recurs << endl;
#endif
   dp = opendir(path);
   if(dp == NULL)
	 {
	 perror("list_files()");
	 return 0;
	 }

   recurs++;
   dbuf = readdir(dp);
   count = 0;
   while( (dbuf != NULL) && (count < len) )
	 {
	 sprintf(nbuf,"%s/%s",path,dbuf->d_name);
#ifdef DEBUG_ON
	 cerr << "\tlist_files(): "
		 << nbuf << endl;
#endif
	 // Check for hidden files [e.g. '.', '..', etc....]
	 if(dbuf->d_name[0]=='.')
	    ok = 0;
	 else
	    ok = stat(nbuf,&sbuf) + 1;  //typ==0 on success, -1 on err.
	 // Check file type.
	 if(S_ISREG(sbuf.st_mode) && ok)
	    {
	    for(i=0; i<IOBUFLEN; i++)
		  listbuf[count][i] = nbuf[i];
	    count++;
	    }
	 else if( (S_ISDIR(sbuf.st_mode) || S_ISLNK(sbuf.st_mode))
			&& ok)			
	    {
	    count += list_files(nbuf,&(listbuf[count]),len-count);
	    }
	 //
	 dbuf = readdir(dp);
	 }
   closedir(dp);
   recurs--;
#ifdef DEBUG_ON
   cerr << "\tlist_files(): return=" << count << endl;
#endif
   return count;
}




//||||||||||||||||||||||||||||||||||||||||||||||||
//
//  Class TiKanDB
//
//||||||||||||||||||||||||||||||||||||||||||||||||



// Constructor for the TiKanDB object.  Parent
// AudDB(void) called first.
//
TiKanDB::TiKanDB(void)
{
   int i;

   pathpos = NULL;
   tmplen = 0;
   pathlst = new char *[NPATH_MAX];
   for(i=0; i<NPATH_MAX; i++)
	 pathlst[i] = new char [IOBUFLEN];
}


// Overloaded constructor for the TiKanDB object.  Parent
// AudDB(void) called first.
//
TiKanDB::TiKanDB(char *fpath) : AudDB(fpath)
{
   int i;

   pathpos = NULL;
   tmplen = 0;
   pathlst = new char *[NPATH_MAX];
   for(i=0; i<NPATH_MAX; i++)
	 pathlst[i] = new char [IOBUFLEN];
   open();
}


// Destructor for the TiKanDB object.
//
TiKanDB::~TiKanDB(void)
{
   int i;

   for(i=0; i<NPATH_MAX; i++)
	 delete [] pathlst[i];
   delete [] pathlst;
   if( (pathpos != NULL) && (tmplen != 0) )
	 {
	 for(i=0; i<tmplen; i++)
	    delete [] pathpos[i];
	 delete [] pathpos;
	 }
}


// Other methods.
//
void TiKanDB::open(void)
{
   char *tptr,pbuf[IOBUFLEN];
   int i,netto;

   // Begin this mess by splitting up the path list.
   strncpy(pbuf,db_name,IOBUFLEN-1);
   pbuf[IOBUFLEN-1] = '\0';
   npaths=0;
   tptr = strtok(pbuf,":");
   while( (tptr != NULL) && (npaths < NPATH_MAX) )
	 {
	 strncpy(pathlst[npaths],tptr,IOBUFLEN-1);
	 pathlst[npaths][IOBUFLEN-1] = '\0';
	 npaths++;
	 tptr = strtok(NULL,":");
	 }

   // Okay, now we scan the paths to count the # of files.
   for(i=0, len_magic=0; i<npaths; i++)
	 len_magic += count_files(pathlst[i]);

   // Next, we build the file list.
   tmplen = len_magic;
   pathpos = new char *[len_magic];
   for(i=0; i<len_magic; i++)
	 pathpos[i] = new char[IOBUFLEN];

   for(i=0, netto=0; i<npaths; i++)
	 netto += list_files(pathlst[i],&(pathpos[netto]),
					 len_magic-netto);
   if(netto < len_magic)
	 len_magic = netto;

   // Lastly, we store the magic numbers, which are just the file
   // names.
   if(magiclist != NULL)
	 delete [] magiclist;
   magiclist = new unsigned long[len_magic];
   for(i=0; i<len_magic; i++)
	 {
	 tptr = strrchr(pathpos[i],'/');
	 if(tptr == NULL)
	    magiclist[i] = 0;
	 else
	    sscanf(tptr+1,"%lx",&(magiclist[i]));
	 }
#ifdef DEBUG_ON
   cerr << "%dbg%\tTiKanDB::open():  " << endl;
   for(i=0; i<len_magic; i++)
	 cerr << "\tfile[" << i << "]=" << pathpos[i] << endl;
   cerr << "\tTotal count: " << len_magic << endl;
#endif
}


// Reads the database entry at the idx^th position in the internal db
// list.
//
bool TiKanDB::read_i(Catalog &ct, int idx)
{
   bool iostat;
   ifstream ifs;

   if( (idx < 0) || (idx >= len_magic) )
	 return false;

#ifdef DEBUG_ON
   cerr << "%dbg%\tTiKanDB::read_i():  file='" 
	   << pathpos[idx] << "'" << endl;
#endif
   ifs.open( pathpos[idx], ios::in | ios::nocreate | ios::binary );
   if (!ifs)
	 {
	 cerr << "TiKanDB::read_i() - Cannot open database file, '"
		 << db_name << "'." << endl;
	 return false;
	 }
   iostat = readEntry(ifs,ct);
   ifs.close();
   return iostat;
}


// Writes a catalog to the database.
//
void TiKanDB::write(Catalog &ct, int &oldlen)
{
   char **olpp;
   int i,j,idx;
   bool iostat;
   ofstream ofs;

   // See if we're doing an append.  If so, lengthen the list and
   // increment len_magic.
   idx = find(ct);
   if(idx < 0)
	 {
#ifdef DEBUG_ON
	 cerr << "%dbg%\tTiKanDB::write(): append. " << endl;
#endif
	 olpp = pathpos;
	 pathpos = new char *[len_magic+1];
	 for(i=0; i<len_magic; i++)
	    {
	    pathpos[i] = new char[IOBUFLEN];
	    for(j=0; j<IOBUFLEN; j++)
		  pathpos[i][j] = olpp[i][j];
	    }
	 if(olpp != NULL)
	    {
	    for(i=0; i<tmplen; i++)
		  delete [] olpp[i];
	    delete [] olpp;
	    }
	 idx = len_magic;
	 pathpos[idx] = new char[IOBUFLEN];
	 sprintf(pathpos[idx],"%s/%08lx",DEFPATH,ct.magic());
	 len_magic++;
	 tmplen = len_magic;
#ifdef DEBUG_ON
	 cerr << "%dbg%\tNew file: " << pathpos[idx] << endl;
#endif
	 }

   ofs.open( pathpos[idx], ios::out | ios::binary );
   if (!ofs)
	 {
	 cerr << "TiKanDB::write() - Cannot open database file, '"
		 << db_name << "'." << endl;
	 return;
	 }
   iostat = writeEntry(ofs,ct);
   ofs.close();
   oldlen = len_magic;
   return;
}


// Reads an Entry from the stream.
//
bool TiKanDB::readEntry(ifstream &ifs, Catalog &ct)
{
   int i,plct;
   int pltmp[LINEBUF];
   char buf[LINEBUF],ttitle[IOBUFLEN];
   char *eptr,*tptr;

   // Read and discard the comments.
   do
	 {
	 ifs.getline(buf,LINEBUF);
	 }
   while(buf[0] == '#');

   // Get the magic number and ignore
   eptr = getEntry(buf,"DISCID");

   // This is tricky - we need to split this entry into artist and
   // title.
   ifs.getline(buf,LINEBUF);
   eptr = getEntry(buf,"DTITLE");
   if(eptr == NULL)
	 {
	 return false;
	 }
   else
	 {
	 tptr=strstr(eptr," / ");
	 tptr[0]='\0';
	 tptr += 3;
	 ct.setArtist(eptr);
	 ct.setTitle(tptr);
	 }

   // Now read the track titles.
   for(i=0; i<ct.len(); i++)
	 {
	 sprintf(ttitle,"TTITLE%d",i);
	 ifs.getline(buf,LINEBUF);
	 eptr = getEntry(buf,ttitle);
	 if(eptr == NULL)
	    return false;
	 ct.setName(i,eptr);
	 }

   // We ignore everything else up to the playlist.
   do
	 {
	 ifs.getline(buf,LINEBUF);
	 eptr = getEntry(buf,"PLAYORDER",false);
	 }
   while( (eptr == NULL) && ifs.good());
   if( (eptr == NULL) || (eptr[0]=='\0') )
	 {
	 cerr << "TiKanDB::readEntry(): Could not find 'PLAYORDER' field"
		 << endl;
	 cerr << "or field empty." << endl;
	 return false;
	 }


   // Now we read in the default playlist.
   plct=0;
   tptr=strtok(eptr,",");
   while( (tptr != NULL) && (plct<LINEBUF) )
	 {
	 sscanf(tptr,"%d",&(pltmp[plct]));
	 plct++;
	 tptr=strtok(NULL,",");
	 }
   if(plct == 0)
	 {
	 ct.resetPlaylist();
	 return true;
	 }
   if((ct.pLen > 0) && (ct.playlist != NULL) )
	 delete [] ct.playlist;
   ct.pLen=plct;
   ct.playlist = new int[plct];
   for(i=0; i<plct; i++)
	 ct.playlist[i] = pltmp[i]-1;

   return true;
}


// Writes a catalog to the stream.
//
bool TiKanDB::writeEntry(ofstream &ofs, Catalog &ct)
{
   bool badness;
   int i,clen;

   // Write file header
   clen = ct.len();
   badness = false;
   ofs << "# xmcd CD database file." 
	  << endl;
   badness |= (!ofs);
   ofs << "# xmcd database format Copyright (C) 1995 Ti Kan"
	  << endl;
   badness |= (!ofs);

   // Write track info into comment.
   ofs << "#" << endl << "# Track frame offsets:" << endl;
   badness |= (!ofs);
   for(i=0; i<clen; i++)
	 {
	 ofs << "#\t" << ct.time(i) << endl;
	 badness |= (!ofs);
	 }
   ofs << "#" << endl << "#  Disc length: " 
	  << ct.time(clen)/FRAME_PER_SEC << " seconds" 
	  << endl;
   badness |= (!ofs);

   // Write magic number, artist, and title.
   ofs << "#" << endl << "DISCID=" << flush;
   badness |= (!ofs);
   ofs << hex << setfill('0') << setw(8) << ct.magic() 
	  << dec << setfill(' ') << endl;
   badness |= (!ofs);
   ofs << "DTITLE=" << ct.artist() << " / " 
	  << ct.title() << endl;
   badness |= (!ofs);

   // Write track names.
   for(i=0; i<clen; i++)
	 {
	 ofs << "TTITLE" << i << "=" << ct.name(i) << endl;
	 badness |= (!ofs);
	 }

   // Write headers for extra info.  Entries are empty
   ofs << "EXTD=  " << endl;
   badness |= (!ofs);
   for(i=0; i<clen; i++)
	 {
	 ofs << "EXTT" << i << "=  " << endl;
	 badness |= (!ofs);
	 }

   // Write playlist.
   ofs << "PLAYORDER=" << flush;
   badness |= (!ofs);
   for(i=0; i<ct.pLen-1; i++)
	 {
	 ofs << ct.playlist[i]+1 << ",";
	 badness |= (!ofs);
	 }
   ofs << ct.playlist[ct.pLen-1]+1 << endl;
   badness |= (!ofs);
   
   if(badness)
	 {
	 cerr << "TiKanDB::writeEntry() - write failure."
		 << "Entry mangled." << endl;
	 cerr << "[You may be able to fix the problem by editing '"
		 << db_name << "' by hand.]" << endl;	 
	 return false;
	 }

   return true;
}



//||||||||||||||||||||||||||||||||
//
// End
//
